DAT programName         byte "DemoFourMotorsAndEncoders130205ba", 0     
CON
{{
  ******* Public Notes *******

  Program written by Duane Degn
  See date code in file name for last update.
  (Example: 130205xx indicates the last update
  was on February 5, 2013.)
  The encoder object used was written by Duane Degn
  based on information given in a Nuts & Volts article
  written by Jon McPhalen.
  The motor control object is based on a two motor
  control object written by Hanno and includes his
  copyright notice.
  
  See child objects for additional credit information.
  
  Program to demonstrate the use of motor and
  encoder drivers.

  To make it easy to use this program with multiple
  robots, I've moved pin assignments to a header
  file.
  The header file also includes other robot specific
  data.

  The user interface of this program is currently a
  bit awkward.
  To change a motor's speed, the motor number needs
  to entered into the terminal window.
  Since the motor ID number will be a single digit,
  the enter key should not be pressed after entering
  the motor ID.
  If the enter key is pressed after entering the motor
  ID, the motor's speed will be set to zero (let's
  call this a "feature").
  After the motor ID has been entered, the program
  will wait for the speed to be entered.
  The enter key will indicate to the program that the
  final digit of the speed has been entered.
  Depending on the motor being used, and your
  power supply, it may be advisable to "ramp" the
  speed of the motor.

  For example when I want to test one of my motors at
  full speed (1000), I initially enter the speed of 500
  to keep the motor from drawing too much current and
  reseting the Propeller. I then increment the speed
  by 100 until I reach the desired speed.

  I'm working on improvements that will automatically ramp
  the speed and adjust PWM levels based on speed information
  from the encoders.

  I believe this program currently uses four of the Propeller's
  eight cogs.
  
  Known Issues:
  
  There are some glitches in the debug display that haven't
  been fixed. I'm not sure if I'm going to bother to fix these
  since the current debug display is temporary.

  I have not calculated the PWM frequency.
  
}}
{
  ******* Private Notes *******

  130125a Use debug statements in parent object
  instead of child object.
  DemoFourMotorsAndEncoders130125a modified from
  DemoFourMotors130125a.
  DemoFourMotorsAndEncoders130125a add encoders.
  31c Comment out motor control.
  0201a Switch to Vex Treads robot.
  0203a Progress! Motor #0 at +1000 target speeds reads
  -945 speed from encoders.
  05ba Fork version 130203b Cleanup older version to post
  to forum. Add more comments.
} 
  _CLKMODE = XTAL1 + PLL16X 
  _XINFREQ = 5_000_000
  
  MAXSPEED = 5000
    
 
  MOTORS_IN_USE = Encoders#TOTAL_ENCODERS

  ' state enumerate 
  #0, WAITING_FOR_MOTOR_ID, FIRST_DIGIT_SPEED, WAITING_FOR_SPEED
  ''(See page 87 of Propeller Manual v1.2 for information on enumberation.)
  

  ' oldDataType enumeration
  #0, NO_OLD_DATA, MOTOR_ONLY_OLD_DATA, MOTOR_AND_T_SPEED_OLD_DATA

  ' newDataType enumeration
  #0, NO_NEW_DATA, MOTOR_ONLY_NEW_DATA, WAITING_T_SPEED_NEW_DATA, PARTIAL_T_SPEED_NEW_DATA, {
    } FULL_T_SPEED_NEW_DATA, TIMED_ERROR_NEW_DATA
  
CON
'' Constants used to format debug information in terminal window.

  TITLE_X = 3                   ' location for title
  TITLE_Y = 3
  DISPLAY_X = TITLE_X              
  DISPLAY_Y = TITLE_Y + 3
  TARGET_SPEED_X = DISPLAY_X 
  TARGET_SPEED_Y = DISPLAY_Y 

  CURRENT_SPEED_X = TARGET_SPEED_X
  CURRENT_SPEED_Y = TARGET_SPEED_Y + 6
  ENCODER_X = CURRENT_SPEED_X
  ENCODER_Y = CURRENT_SPEED_Y + 6
  PIN_STATES_X = ENCODER_X
  PIN_STATES_Y = ENCODER_Y + 6  
  OLD_INFO_X = PIN_STATES_X
  OLD_INFO_Y = PIN_STATES_Y + 3
  REQUEST_INPUT_X = OLD_INFO_X 
  REQUEST_INPUT_Y = OLD_INFO_Y + 2
  NEW_INFO_X = REQUEST_INPUT_X
  NEW_INFO_Y = REQUEST_INPUT_Y + 2

  SPEED_BUFFER_X = NEW_INFO_X 
  SPEED_BUFFER_Y = NEW_INFO_Y + 3
  PID_X = DISPLAY_X             ' PID information is not currently used.
  PID_Y = DISPLAY_Y + 9
  ADJUSTED_PID_X = PID_X
  ADJUSTED_PID_Y = PID_Y + 4

  NEW_DATA_DISPLAY_TIME = 10    ' Keep this less than half the rollover time. 
 

VAR

  long enablePin[MOTORS_IN_USE], motorPinRed[MOTORS_IN_USE], motorPinBlack[MOTORS_IN_USE]
  long speedPtr, arrayPtr, previousTSpeed[MOTORS_IN_USE], frozenTSpeed[MOTORS_IN_USE]
  long encoder[MOTORS_IN_USE]
  long bufferSpeed[MOTORS_IN_USE]
  long tempSpeed, speedSign, previousTargetSpeed[MOTORS_IN_USE], previousMotorIndex
  long newDataTimer, newDataInterval
  long frozenCSpeed[MOTORS_IN_USE], previousCSpeed[MOTORS_IN_USE]
  long frozenEncoder[MOTORS_IN_USE], previousEncoder[MOTORS_IN_USE]
  long previousIna, frozenIna, frozenBuffer[Encoders#SPEED_BUFFER_SIZE * Encoders#TOTAL_ENCODERS]
  long bufferPtr
  byte state, motorIndex, oldDataType, newDataType
   
OBJ

  Pst : "Parallax Serial Terminal"                      ' Uses one cog
  Pwm : "fourMotorsOneCog130205a"                       ' Uses one cog
  Encoders : "FourQuadratureEncoders130205a"            ' Uses one cog
  Header : "HeaderVexTreads130201a"                     
  'Header : "HeaderRover5_130130a"
  
PUB Setup

  Pst.Start(115200)
  
  enablePin[0] := Header#WHEEL_ENABLE_0
  motorPinRed[0] := Header#WHEEL_RED_0
  motorPinBlack[0] := Header#WHEEL_BLACK_0
  enablePin[1] := Header#WHEEL_ENABLE_1
  motorPinRed[1] := Header#WHEEL_RED_1 
  motorPinBlack[1] := Header#WHEEL_BLACK_1 
  enablePin[2] := Header#WHEEL_ENABLE_2
  motorPinRed[2] := Header#WHEEL_RED_2
  motorPinBlack[2] := Header#WHEEL_BLACK_2
  enablePin[3] := Header#WHEEL_ENABLE_3
  motorPinRed[3] := Header#WHEEL_RED_3 
  motorPinBlack[3] := Header#WHEEL_BLACK_3 
  
  Encoders.StartEncoders(Header#FIRST_ENCODER_0, @encoder, @bufferSpeed)
  
  Pwm.Start(motorPinRed[0], motorPinRed[1], motorPinRed[2], motorPinRed[3], {
  } motorPinBlack[0], motorPinBlack[1], motorPinBlack[2], motorPinBlack[3], {
  } enablePin[0], enablePin[1], enablePin[2], enablePin[3])
  
  arrayPtr := Pwm.GetArrayLocation
  speedPtr := Pwm.GetSpeedLocation
  bufferPtr := Encoders.GetBufferLocation
  oldDataType := NO_OLD_DATA
  newDataType := NO_NEW_DATA 
  newDataInterval := clkfreq * NEW_DATA_DISPLAY_TIME
   
  MainLoop

PUB MainLoop  

  repeat
 
    result := Pst.RxCount
    if result
      if state == WAITING_FOR_MOTOR_ID   
        DisplayMenuSpeed
      elseif state == FIRST_DIGIT_SPEED or state == WAITING_FOR_SPEED
        ReceiveSpeedInput
    CheckNewData
    DisplayTargetSpeed
    DisplayPresentSpeed
    DisplayEncoder
    DisplayPinStates
    DisplaySpeedBuffer
      
PRI ReceiveSpeedInput | localCharacter
'' Examine input character and adjust speed as needed.

  localCharacter := Pst.CharIn
  if state == FIRST_DIGIT_SPEED    
    if localCharacter == "-"
      speedSign := -1
      state := WAITING_FOR_SPEED
      return
    elseif localCharacter == "+"
      speedSign := 1
      state := WAITING_FOR_SPEED
      return
  if localCharacter < "0" or localCharacter > "9"
    if localCharacter == 13
      ''Finalize speed input
      tempSpeed *= speedSign
      Pst.Position(OLD_INFO_X, OLD_INFO_Y + 1)
      Pst.str(string("The new speed for motor #"))
      Pst.Dec(motorIndex)      
      Pst.str(string(" is "))
      Pst.Dec(tempSpeed)
      Pst.ClearEnd   
      Pwm.UpdateSingle(motorIndex, tempSpeed)
      ' Below may be used to keep pairs of motors set to
      ' identical speeds.
      {if motorIndex < 2
        Pwm.UpdateSingle(motorIndex + 2, tempSpeed)
      else
        Pwm.UpdateSingle(motorIndex - 2, tempSpeed)
         }
      previousMotorIndex := motorIndex
      DisplayOldData 
      DisplayNewData
      newDataTimer := cnt
      waitcnt(clkfreq / 4 + cnt)
      DisplayMenuMotor
      return
    Pst.Position(NEW_INFO_X, NEW_INFO_Y)
    Pst.str(string("The character ", 34))
    Pst.char(localCharacter)
    Pst.str(string(34, " is not a valid entry."))
    Pst.ClearEnd
    Pst.Position(NEW_INFO_X, NEW_INFO_Y + 1)
    Pst.str(string("Please enter a digit between ", 34, "0", 34, " and ", 34, "9", 34, "."))
    Pst.ClearEnd
    Pst.Position(NEW_INFO_X, NEW_INFO_Y + 2)
    Pst.str(string("Press ", 34, "enter", 34, " to submit number."))
    Pst.ClearEnd
    Pst.Position(NEW_INFO_X, NEW_INFO_Y + 3)
    if state == WAITING_FOR_SPEED
      Pst.str(string("So far you have entered ", 34))
      if tempSpeed == 0
        if speedSign == -1
          Pst.char("-")
        else
          Pst.char("+")
        Pst.ClearEnd  
      else
        Pst.Dec(speedSign * tempSpeed)
      Pst.char(34)
      Pst.ClearEnd      
    else    
      Pst.str(string("You have not entered any digits yet.            "))
      Pst.ClearEnd
    newDataType := TIMED_ERROR_NEW_DATA
    newDataTimer := cnt  
  else
    tempSpeed *= 10
    tempSpeed += localCharacter - "0"
    state := WAITING_FOR_SPEED
    
PRI DisplayMenuMotor
'' Request motor ID input

  Pst.Home
  repeat result from 0 to TITLE_Y - 1
    Pst.Position(0, result)
    Pst.ClearEnd
  Pst.Position(TITLE_X, TITLE_Y)
  Pst.str(@textTitle)
  Pst.Position(TITLE_X, TITLE_Y + 1)
  Pst.str(string("programName = "))
  Pst.str(@programName)
  Pst.ClearEnd
  Pst.Position(REQUEST_INPUT_X, REQUEST_INPUT_Y)
  Pst.str(@menuMotor0)
  Pst.ClearEnd
  Pst.Position(REQUEST_INPUT_X, REQUEST_INPUT_Y)
  Pst.str(@menuMotor1)
  Pst.ClearEnd
  state := WAITING_FOR_MOTOR_ID
  tempSpeed := 0
  speedSign := 1
  longfill(@previousTSpeed, -1, MOTORS_IN_USE)           ' Force speed display
  longfill(@previousCSpeed, -1, MOTORS_IN_USE)           
  longfill(@previousEncoder, -1, MOTORS_IN_USE)
  
PRI DisplayMenuSpeed | localCharacter
'' Receive motor ID input.
'' Request speed input.

  localCharacter := Pst.CharIn
  ifnot localCharacter          ' return if null character is received
    return
  localCharacter -= "0"
  Pst.Position(OLD_INFO_X, OLD_INFO_Y)
  Pst.str(string("You selected motor ID #"))
  Pst.dec(localCharacter)
  Pst.ClearEnd
  
  if localCharacter => 0 and localCharacter =< MOTORS_IN_USE
    motorIndex := localCharacter
    Pst.Position(REQUEST_INPUT_X, REQUEST_INPUT_Y)
    Pst.str(string("Please enter new speed for motor #"))
    Pst.dec(motorIndex)
    Pst.ClearEnd
    Pst.Position(REQUEST_INPUT_X, REQUEST_INPUT_Y + 1)
    Pst.str(string("(-1000 to 1000)                    "))
    Pst.ClearEnd
    state := FIRST_DIGIT_SPEED    
  else
    Pst.Position(NEW_INFO_X, NEW_INFO_Y)
    Pst.str(string("Not a valid motor ID value."))
    Pst.ClearEnd
    newDataType := TIMED_ERROR_NEW_DATA
    newDataTimer := cnt
    DisplayMenuMotor
 
PRI DisplayOldData

  Pst.Position(OLD_INFO_X, OLD_INFO_Y)
  case oldDataType
    NO_OLD_DATA:
      Pst.ClearEnd
      Pst.NewLine
      Pst.ClearEnd
    MOTOR_ONLY_OLD_DATA:
      Pst.str(@oldData0)
      Pst.dec(previousMotorIndex)
      Pst.ClearEnd
    MOTOR_AND_T_SPEED_OLD_DATA:
      Pst.str(@oldData0)
      Pst.dec(previousMotorIndex)
      Pst.ClearEnd
      Pst.Position(OLD_INFO_X, OLD_INFO_Y + 1)
      Pst.str(@oldData1)
      Pst.dec(previousTargetSpeed)
      Pst.ClearEnd
  
PRI DisplayNewData
'' Information displayed in terminal window
'' for a limited amount of time.

  Pst.Position(NEW_INFO_X, NEW_INFO_Y)
  case newDataType
    NO_NEW_DATA:                
      Pst.str(string("No New Data")) 
      Pst.ClearEnd
      Pst.Position(NEW_INFO_X, NEW_INFO_Y + 1)
      Pst.ClearEnd
     
    MOTOR_ONLY_NEW_DATA:
      Pst.str(@newData0)
      Pst.dec(motorIndex)
      Pst.ClearEnd
    
    PARTIAL_T_SPEED_NEW_DATA:
      Pst.str(@newData0)
      Pst.dec(motorIndex)
      Pst.ClearEnd
      Pst.Position(NEW_INFO_X, NEW_INFO_Y + 1)
      Pst.str(string("targetSpeed["))
      Pst.ClearEnd
   
    FULL_T_SPEED_NEW_DATA:
      Pst.str(@newData0)
      Pst.dec(motorIndex)
      Pst.ClearEnd
      Pst.Position(NEW_INFO_X, NEW_INFO_Y + 1)
      Pst.str(@newData1)
      Pst.dec(previousTargetSpeed)
      Pst.ClearEnd
    
    TIMED_ERROR_NEW_DATA:
      return
      
PRI CheckNewData
'' Limit display time of some data.

  if newDataType <> NO_NEW_DATA 
    if cnt - newDataInterval > newDataTimer
      newDataType := NO_NEW_DATA
      DisplayNewData
  
PRI DisplayTargetSpeed
'' Diplay dbug information.

  Pst.Position(TARGET_SPEED_X, TARGET_SPEED_Y - 1)
  Pst.str(@textTSpeed)
  'waitcnt(clkfreq * 4 + cnt)
  longmove(@frozenTSpeed, speedPtr, MOTORS_IN_USE)
  repeat result from 0 to MOTORS_IN_USE - 1
    Pst.Position(TARGET_SPEED_X, TARGET_SPEED_Y + result)
    if frozenTSpeed[result] <> previousTSpeed[result]
      Pst.str(string("targetSpeed["))
      Pst.dec(result)
      Pst.str(string("] = "))
      Pst.dec(frozenTSpeed[result] / 4)
      Pst.ClearEnd
      previousTSpeed[result] := frozenTSpeed[result]
      
PRI DisplayPresentSpeed
'' Diplay dbug information.

  Pst.Position(CURRENT_SPEED_X, CURRENT_SPEED_Y - 1)
  Pst.str(@textCSpeed)
  'waitcnt(clkfreq * 4 + cnt)
  
  longmove(@frozenCSpeed, @bufferSpeed, MOTORS_IN_USE)
  repeat result from 0 to MOTORS_IN_USE - 1
    Pst.Position(CURRENT_SPEED_X, CURRENT_SPEED_Y + result)
    'if frozenCSpeed[result] <> previousCSpeed[result]
      Pst.str(string("currentSpeed["))
      Pst.dec(result)
      Pst.str(string("] = "))
      Pst.dec(frozenCSpeed[result])
      Pst.ClearEnd
      previousCSpeed[result] := frozenCSpeed[result]
      
PRI DisplayEncoder
'' Diplay dbug information.

  Pst.Position(ENCODER_X, ENCODER_Y - 1)
  Pst.str(@textEncoder)
  'waitcnt(clkfreq * 4 + cnt)
  
  longmove(@frozenEncoder, @encoder, MOTORS_IN_USE)
  repeat result from 0 to MOTORS_IN_USE - 1
    Pst.Position(ENCODER_X, ENCODER_Y + result)
    'if frozenEncoder[result] <> previousEncoder[result]
      Pst.str(string("encoder["))
      Pst.dec(result)
      Pst.str(string("] = "))
      Pst.dec(frozenEncoder[result])
      Pst.ClearEnd
      previousEncoder[result] := frozenEncoder[result]

PRI DisplayPinStates
'' Diplay dbug information.

  frozenIna := ina
  if frozenIna <> previousIna
    Pst.Position(PIN_STATES_X, PIN_STATES_Y)
    Pst.str(@textPinStates)
    Pst.bin(frozenIna, 32)
    Pst.ClearEnd
    previousIna := frozenIna

PRI DisplaySpeedBuffer | localIndex
'' This method dumps the buffer the child object uses
'' to keep track of speed.
'' Generally this information isn't needed, but it's interesting
'' to see while debugging.

  Pst.Position(SPEED_BUFFER_X, SPEED_BUFFER_Y - 1)
  Pst.str(@textBuffer)
  'waitcnt(clkfreq * 4 + cnt)

  longmove(@frozenBuffer, bufferPtr, Encoders#SPEED_BUFFER_SIZE * Encoders#TOTAL_ENCODERS)
  repeat result from 0 to Encoders#TOTAL_ENCODERS - 1
    Pst.Position(SPEED_BUFFER_X, SPEED_BUFFER_Y + result)   
    Pst.str(string("buffer["))
    Pst.dec(result)
    Pst.str(string("] = "))
    repeat localIndex from 0 to Encoders#SPEED_BUFFER_SIZE - 2
      Pst.dec(frozenBuffer[(result * Encoders#SPEED_BUFFER_SIZE) + localIndex])
      Pst.char(",")
      Pst.char(32)
    Pst.dec(frozenBuffer[(result * Encoders#SPEED_BUFFER_SIZE) + Encoders#SPEED_BUFFER_SIZE - 1])  
    Pst.ClearEnd

DAT textTitle byte "Four Motors with Quadrature Encoders Demo", 0

menuMotor0    byte "Please enter ID number of motor", 0
menuMotor1    byte " to change its speed.", 0


textTSpeed    byte "Target Speeds:", 0

textCSpeed    byte "Current Speeds:", 0

textEncoder   byte "Current Encoder Readings:", 0

textPinStates byte "Ina = %", 0

textBuffer    byte "Speed Averaging Buffers:", 0

oldData0      byte "Last Motor Changed: #", 0
oldData1      byte "Last Speed Input: ", 0

newData0      byte "Motor Selected: #", 0
newData1      byte "Speed Entered = ", 0

DAT '' License
{{
*********************************************************************************************
*                             TERMS OF USE: MIT License                                     *                                                                   
*********************************************************************************************
* Permission is hereby granted, free of charge, to any person obtaining a copy of this      *
* software and associated documentation files (the "Software"), to deal in the Software     *
* without restriction, including without limitation the rights to use, copy, modify,        *
* merge, publish, distribute, sublicense, and/or sell copies of the Software, and to        *
* permit persons to whom the Software is furnished to do so, subject to the following       *
* conditions:                                                                               *
*                                                                                           *                                    
* The above copyright notice and this permission notice shall be included in all copies or  *
* substantial portions of the Software.                                                     *
*                                                                                           *                                   
* The software is provided "as is", without warranty of any kind, express or implied,       *
* including but not limited to the warranties of merchantability, fitness for a particular  *
* and noninfringment. In no event shall the authors or copyright holders be liable for any  *
* claim, damages or other liability, whether in an action of contract, tort or otherwise,   *
* arising from, of or in connection with the software of the use of other dealings in the   *
* software.                                                                                 *
*********************************************************************************************
}}            