DAT programName byte "FourQuadratureEncoders130205a", 0
CON
{{
  ******* Public Notes *******

  This program is used to monitor four quadrature
  encoders like those used by the Rover 5 chassis.
  See comments at start of each method.

  This object was written by Duane Degn

  The code based on a the encoder technique shown in a
  Nuts & Volts article by Jon "JonnyMac" McPhalen.
  
}}
{
  ******* Private Notes *******
  
  FourQuadratureEncoders130125a modifed from VexMecanumRc121219b
  130125a Start removing non-encoder code from program and
  make this object a child object.
  0205a Add additional comments in preperation to posting to forum.
  

}
CON

  TOTAL_ENCODERS = 4            ' This is not user changable. The other constants may
                                ' adjusted by user.
  SPEED_BUFFER_SIZE = 16
  CYCLES_PER_ENCODER_CHECK = 800
  ENCODER_LOOPS_PER_WRITE = 512
  
VAR

  long speedBuffer[SPEED_BUFFER_SIZE * TOTAL_ENCODERS] 
  long debugTime
            
PUB StartEncoders(firstEncoderPin, encoderPtr, speedPtr)
'' This method expects encoder pins to be sequential from
'' "firstEncoderPin".
'' The value "encoderPtr" should be the memory location
'' of four sequential longs where incremented encoder values
'' will be written.
'' The value "speedPtr" should be the memory location
'' of four sequential longs where the sum of "speedBuffer" values
'' will be written.
'' This method needs to be called to start using object.
'' There isn't a "Stop" method so multiple calls
'' will launch multiple cogs which will likely cause serious
'' problems for parent program.

  encoderPin0 := firstEncoderPin 
  encoderPin1 := firstEncoderPin + 2 
  encoderPin2 := firstEncoderPin + 4
  encoderPin3 := firstEncoderPin + 6
  timeAddress := @debugTime
  buffer0 := @speedBuffer       ' first location in each buffer (static)
  buffer1 := buffer0 + (SPEED_BUFFER_SIZE * 4)
  buffer2 := buffer1 + (SPEED_BUFFER_SIZE * 4)
  buffer3 := buffer2 + (SPEED_BUFFER_SIZE * 4)
  
  buffer0Head := buffer0        ' current location in each buffer (dynamic)
  buffer1Head := buffer1
  buffer2Head := buffer2
  buffer3Head := buffer3
   
  speedBufferSumPtr0 := speedPtr
  speedBufferSumPtr1 := speedBufferSumPtr0 + 4
  speedBufferSumPtr2 := speedBufferSumPtr1 + 4
  speedBufferSumPtr3 := speedBufferSumPtr2 + 4
  result := cognew(@enter, encoderPtr)

PUB GetFileName
'' This method returns the name of this object.
'' This method is not needed for proper operation of object
'' and may be deleted.

  result := @programName

PUB GetDebugTime
'' This method returns the value of "debugTime" to aid
'' in debugging program.
'' This method is not needed for proper operation of object
'' and may be deleted.

  result := debugTime
  
PUB GetBufferLocation
'' This method returns the location of "speedBuffer" to aid
'' in debugging program.
'' This method is not needed for proper operation of object
'' and may be deleted.

  result := @speedBuffer
  
DAT                     org       0

enter                   mov     encoderAddress, par 
                        mov     encoderAddress0, encoderAddress 
                        mov     encoderAddress1, encoderAddress0
                        add     encoderAddress1, #4 
                        mov     encoderAddress2, encoderAddress1
                        add     encoderAddress2, #4 
                        mov     encoderAddress3, encoderAddress2
                        add     encoderAddress3, #4 

                      
                        mov     loopsTilWrite, loopsToWrite                 
                        mov     headCountDown, bufferSize
                        mov     oldscan0, ina                   ' get start-up inputs
                        mov     oldscan1, oldscan0                    
                        mov     oldscan2, oldscan0
                        mov     oldscan3, oldscan0                                            
                        shr     oldscan0, encoderPin0           ' z-align scan   
                        and     oldscan0, #%11                  ' clear other bits
                        shr     oldscan1, encoderPin1           ' z-align scan   
                        and     oldscan1, #%11                  ' clear other bits
                        shr     oldscan2, encoderPin2           ' z-align scan   
                        and     oldscan2, #%11                  ' clear other bits
                        shr     oldscan3, encoderPin3           ' z-align scan
                        and     oldscan3, #%11                  ' clear other bits
                        
                        mov     timer, cnt                      ' sync with system cnt
                        add     timer, enctiming                ' set loop timing                        
                        mov     speedTimer, timer

'-------------------------------------------------------------------------------

encloop                 waitcnt timer, enctiming                ' hold for loop timing
                        wrlong   timer, timeAddress             ' This should be at least
                                                                ' as long as the longest
                                                                ' possible loop will take
                                                                ' 4 + (4 * 30) + (4 * 9) +
                                                                ' 50 = 120 instructions =
                                                                ' 480 clock cycles
scan                    djnz    loopsTilWrite, #continueScan
                        call    #refreshCount
continueScan            mov     originalScan, ina               ' get current inputs

                        mov     newscan, originalScan           
                        shr     newscan, encoderPin0
                        mov     oldScan, oldscan0               
                        mov     encoderTotal, encoderTotal0     
                        mov     encoderAddress, encoderAddress0   
                        call    #chkEncoder                     ' run alrorithm with active variables
                        mov     oldscan0, newscan 
                        mov     encoderTotal0, encoderTotal
                        add     encoderQuanta0, encoderChange
                                                                        
                        mov     newscan, originalScan           
                        shr     newscan, encoderPin1
                        mov     oldScan, oldscan1               
                        mov     encoderTotal, encoderTotal1     
                        mov     encoderAddress, encoderAddress1   
                        call    #chkEncoder                     ' run alrorithm with active variables
                        mov     oldscan1, newscan 
                        mov     encoderTotal1, encoderTotal
                        add     encoderQuanta1, encoderChange

                        mov     newscan, originalScan           
                        shr     newscan, encoderPin2
                        mov     oldScan, oldscan2               
                        mov     encoderTotal, encoderTotal2     
                        mov     encoderAddress, encoderAddress2   
                        call    #chkEncoder                     ' run alrorithm with active variables
                        mov     oldscan2, newscan 
                        mov     encoderTotal2, encoderTotal
                        add     encoderQuanta2, encoderChange

                        mov     newscan, originalScan           
                        shr     newscan, encoderPin3
                        mov     oldScan, oldscan3               
                        mov     encoderTotal, encoderTotal3     
                        mov     encoderAddress, encoderAddress3   
                        call    #chkEncoder                     ' run alrorithm with active variables
                        mov     oldscan3, newscan 
                        mov     encoderTotal3, encoderTotal
                        add     encoderQuanta3, encoderChange

                        jmp     #encloop

'-------------------------------------------------------------------------------
                        
chkencoder              mov     encoderChange, zero
                        and     newscan, #%11                   ' about 30 instructions
                        cmp     newscan, oldScan        wz      ' check for change
              if_z      jmp     #chkencoder_ret 'encloop 'chkencoder ' ' wait if none
                        
case11                  cmp     oldscan, #%11           wz      ' oldscan == %11?         
              if_nz     jmp     #case01                         '20 if no, try next         
                        cmp     newscan, #%01           wz      ' clockwise (postive)? Z = yes 
                        jmp     #update

case01                  cmp     oldscan, #%01           wz
              if_nz     jmp     #case00
                        cmp     newscan, #%00           wz   '40
                        jmp     #update                        

case00                  cmp     oldscan, #%00           wz      
              if_nz     jmp     #case10                         
                        cmp     newscan, #%10           wz      
                        jmp     #update                      '60
                                                                          
case10                  cmp     oldscan, #%10           wz
              if_nz     jmp     #chkencoder_ret                 ' (should never happen)
                        cmp     newscan, #%11           wz

update        if_z      mov     encoderChange, one                ' increment value

decvalue      if_nz     mov     encoderChange, negOne              '80 decrement

                        add     encoderTotal, encoderChange
writeTotal              wrlong  encoderTotal, encoderAddress' 88 to 92update hub
                                               
chkencoder_ret          ret
                       
'-------------------------------------------------------------------------------
refreshCount            mov     loopsTilWrite, loopsToWrite
                        rdlong  oldQuanta, buffer0Head
                        sub     bufferTotal0, oldQuanta
                        add     bufferTotal0, encoderQuanta0
                        wrlong  encoderQuanta0, buffer0Head
                        add     buffer0Head, #4                   ' use #1 if buffer is byte size
                        mov     encoderQuanta0, #0                ' clear the encoder total
                        wrlong  bufferTotal0, speedBufferSumPtr0

                        rdlong  oldQuanta, buffer1Head
                        sub     bufferTotal1, oldQuanta
                        add     bufferTotal1, encoderQuanta1
                        wrlong  encoderQuanta1, buffer1Head
                        add     buffer1Head, #4                   ' use #1 if buffer is byte size
                        mov     encoderQuanta1, #0
                        wrlong  bufferTotal1, speedBufferSumPtr1
 
                        rdlong  oldQuanta, buffer2Head
                        sub     bufferTotal2, oldQuanta
                        add     bufferTotal2, encoderQuanta2
                        wrlong  encoderQuanta2, buffer2Head
                        add     buffer2Head, #4                   ' use #1 if buffer is byte size
                        mov     encoderQuanta2, #0
                        wrlong  bufferTotal2, speedBufferSumPtr2
 
                        rdlong  oldQuanta, buffer3Head
                        sub     bufferTotal3, oldQuanta
                        add     bufferTotal3, encoderQuanta3
                        wrlong  encoderQuanta3, buffer3Head
                        add     buffer3Head, #4                   ' use #1 if buffer is byte size
                        mov     encoderQuanta3, #0
                        wrlong  bufferTotal3, speedBufferSumPtr3
 
                        
                        djnz    headCountDown, #refreshCount_ret
                        call    #refreshHead
refreshCount_ret        ret
'-------------------------------------------------------------------------------
refreshHead             mov     buffer0Head, buffer0
                        mov     buffer1Head, buffer1
                        mov     buffer2Head, buffer2
                        mov     buffer3Head, buffer3   
                        mov     headCountDown, bufferSize
refreshHead_ret         ret

'-------------------------------------------------------------------------------

zero                    long 0
one                     long 1
negOne                  long -1
enctiming               long CYCLES_PER_ENCODER_CHECK '800 '20000 '60_000 '30_000 '3200 '1600 '        '
bufferSize              long SPEED_BUFFER_SIZE
loopsToWrite            long ENCODER_LOOPS_PER_WRITE
'speedInterval           long _SpeedInterval
'timeQuanta              long _SpeedInterval / 512 '255
'maxSpeed                long 511 '255
'minSpeed                long -511 '255
'pGain                   long _PGain
'pGainMask               long 0
'bufferAndMask           long 0
timeAddress             long 0
encoderAddress          long 0
encoderTotal            long 0 
encoderTotal0           long 0 
encoderTotal1           long 0 
encoderTotal2           long 0 
encoderTotal3           long 0
encoderChange           long 0
encoderQuanta0          long 0
encoderQuanta1          long 0
encoderQuanta2          long 0
encoderQuanta3          long 0
previousEncoderTotal0   long 0 
previousEncoderTotal1   long 0 
previousEncoderTotal2   long 0 
previousEncoderTotal3   long 0

encoderPin0             long 0
encoderPin1             long 0
encoderPin2             long 0
encoderPin3             long 0

' first location in each buffer
buffer0                 long 0
buffer1                 long 0
buffer2                 long 0
buffer3                 long 0

' current location in each buffer
buffer0Head             long 0
buffer1Head             long 0
buffer2Head             long 0
buffer3Head             long 0

tempHead                long 0
bufferTotal0            long 0
bufferTotal1            long 0
bufferTotal2            long 0
bufferTotal3            long 0
speedBufferSumPtr0      long 0
speedBufferSumPtr1      long 0
speedBufferSumPtr2      long 0
speedBufferSumPtr3      long 0
  
'bufferHead              long 0
'oldSpeed                long 0
'newSpeed                long 0
'bufferBaseAddress       long 0
'debugCount              long 0
'encoderCount            long 0

'timeBufferOffset        long 0
'bufferPtr               res 1
encoderAddress0         res 1
encoderAddress1         res 1
encoderAddress2         res 1
encoderAddress3         res 1
oldScan                 res 1
oldscan0                res 1
oldscan1                res 1
oldscan2                res 1
oldscan3                res 1
timeCount               res 1
timer                   res 1
speedTimer              res 1
speedTimerTemp          res 1

originalScan            res 1
newscan                 res 1

difference              res 1
oldQuanta               res 1
absoluteTarget          res 1
timePtr                 res 1
headCountDown           res 1
loopsTilWrite           res 1 
              fit
              
DAT '' License
{{
*********************************************************************************************
*                             TERMS OF USE: MIT License                                     *                                                                   
*********************************************************************************************
* Permission is hereby granted, free of charge, to any person obtaining a copy of this      *
* software and associated documentation files (the "Software"), to deal in the Software     *
* without restriction, including without limitation the rights to use, copy, modify,        *
* merge, publish, distribute, sublicense, and/or sell copies of the Software, and to        *
* permit persons to whom the Software is furnished to do so, subject to the following       *
* conditions:                                                                               *
*                                                                                           *                                    
* The above copyright notice and this permission notice shall be included in all copies or  *
* substantial portions of the Software.                                                     *
*                                                                                           *                                   
* The software is provided "as is", without warranty of any kind, express or implied,       *
* including but not limited to the warranties of merchantability, fitness for a particular  *
* and noninfringment. In no event shall the authors or copyright holders be liable for any  *
* claim, damages or other liability, whether in an action of contract, tort or otherwise,   *
* arising from, of or in connection with the software of the use of other dealings in the   *
* software.                                                                                 *
*********************************************************************************************
}}        